#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from yubiauth import settings, YubiAuth
from yubiauth.util import validate_otp
from yubiauth.util.controller import Controller
from yubiauth.util.model import Session
from yubiauth.client.model import UserSession, AttributeType, PERMS

__all__ = [
    'Client',
    'requires_otp'
]

REVOKE_KEY = '_revoke'


def requires_otp(user):
    sl = settings['security_level']
    count = len([key for key in user.yubikeys.values() if key.enabled])
    return not (sl == 0 or (sl == 1 and count == 0))


def authenticate_otp(user, otp):
    if otp:
        if settings['auto_provision'] and len(user.yubikeys) == 0:
            if validate_otp(otp):
                user.assign_yubikey(otp)
                return True
            else:
                return False
        else:
            return user.validate_otp(otp)
    else:
        return not requires_otp(user)


class Client(Controller):
    """
    Main class for accessing user data.
    """
    def __init__(self, session=Session()):
        super(Client, self).__init__(session)
        self.auth = YubiAuth(session)

    def _user_for_otp(self, otp):
        if settings['yubikey_id']:
            yubikey = self.auth.get_yubikey(otp[:-32])
            if yubikey.enabled and len(yubikey.users) == 1:
                return yubikey.users[0]
        raise ValueError("Unable to locate user!")

    def authenticate(self, username, password, otp=None):
        if not username and otp:
            user = self._user_for_otp(otp)
        else:
            user = self.auth.get_user(username)
        if user.validate_password(password):
            if authenticate_otp(user, otp):
                return user
        else:
            # Consume the OTP even if the password was incorrect.
            validate_otp(otp)
        raise ValueError("Invalid credentials!")

    def create_session(self, username, password, otp=None):
        user = self.authenticate(username, password, otp)
        prefix = otp[:-32] if otp else None
        user_session = UserSession(user.name, prefix)
        self.session.add(user_session)
        # Prevent loading the user twice
        user_session._user = user
        return user_session

    def get_session(self, sessionId):
        try:
            user_session = self.session.query(UserSession).filter(
                UserSession.sessionId == sessionId).one()
        except Exception:
            raise ValueError("Session not found!")

        user_session.update_used()

        if user_session.is_expired() or not user_session.user:
            user_session.delete()
            raise ValueError("Session is expired!")

        return user_session

    def clear_sessions(self, user=None):
        query = self.session.query(UserSession)
        if user:
            query = query.filter(UserSession.username == user.name)
        query.delete()

    def create_attribute(self, *args, **kwargs):
        attribute = AttributeType(*args, **kwargs)
        self.session.add(attribute)
        return attribute

    def get_attributes(self):
        return self.session.query(AttributeType).all()

    def generate_revocation(self, prefix):
        yubikey = self.auth.get_yubikey(prefix)
        code = UserSession(REVOKE_KEY).sessionId
        yubikey.attributes[REVOKE_KEY] = code
        return code

    def revoke(self, code):
        kwargs = {REVOKE_KEY: code}
        keys = self.auth.query_yubikeys(**kwargs)
        if not len(keys) == 1:
            raise ValueError('Invalid revocation code!')
        keys[0].enabled = False

    def sign_up(self, username, password, otp=None, attributes={}):
        validate_attributes(self.get_attributes(), attributes)

        if otp and not validate_otp(otp):
            raise ValueError('Invalid OTP!')

        user = self.auth.create_user(username, password)
        user.attributes.update(attributes)
        if otp:
            user.assign_yubikey(otp)


def validate_attributes(user_attrs, supplied_attrs, perm_level=PERMS['USER']):
    for attr in user_attrs:
        key = attr.key
        if key in supplied_attrs:
            if attr.edit_perms > perm_level:
                raise ValueError('Not permitted to set attribute: %s' % key)
            if not attr.validate(supplied_attrs[key]):
                raise ValueError('Invalid value for attribute: %s = "%s"'
                                 % (key, supplied_attrs[key]))
        elif attr.required:
            raise ValueError('Missing required attribute: %s' % key)



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from wsgiref.simple_server import make_server
from webob import exc
from webob.dec import wsgify

from yubiauth.core.rest import application as core_api
from yubiauth.client.rest import application as client_api


class YubiAuthAPI(object):
    def __init__(self):
        self._apis = [
            core_api,
            client_api
        ]

    @wsgify
    def __call__(self, request):
        for api in self._apis:
            if request.path.startswith(api._base_path):
                return api(request)

        raise exc.HTTPNotFound


application = YubiAuthAPI()

if __name__ == '__main__':
    httpd = make_server('localhost', 8080, application)
    httpd.serve_forever()



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from wsgiref.simple_server import make_server
from yubiauth.util import validate_otp
from yubiauth.util.rest import (REST_API, Route, json_response, json_error,
                                extract_params)
from yubiauth import settings
from yubiauth.client.controller import Client

import logging as log

SESSION_COOKIE = 'YubiAuth-Session'
SESSION_HEADER = 'X-%s' % SESSION_COOKIE
REVOKE_ATTRIBUTE = '_REVOKE'


def require_session(func):
    def inner(self, request, *args, **kwargs):
        if not request.session:
            return json_error('Session required!')

        return func(self, request, *args, **kwargs)
    return inner


class ClientAPI(REST_API):
    __routes__ = [
        Route(r'^login$', 'login'),
        Route(r'^authenticate$', 'authenticate'),
        Route(r'^logout$', 'logout'),
        Route(r'^status$', 'status'),
        Route(r'^password$', post='change_password'),
        Route(r'^yubikey$', post='assign_yubikey'),
        Route(r'^revoke/generate$', 'generate_revocation'),
        Route(r'^revoke$', post='revoke_yubikey')
    ]

    def _call_setup(self, request):
        request.client = Client()
        request.session = None
        if SESSION_COOKIE in request.cookies:
            sessionId = request.cookies[SESSION_COOKIE]
        elif SESSION_HEADER in request.headers:
            sessionId = request.headers[SESSION_HEADER]

        try:
            request.session = request.client.get_session(sessionId)
        except:
            pass

    def _call_teardown(self, request, response):
        try:
            if request.session:
                sessionId = str(request.session.sessionId)
                # TODO: Roll session key?
                if SESSION_COOKIE in request.cookies and \
                        request.cookies[SESSION_COOKIE] == sessionId:
                    return
                https = request.scheme == 'https'
                response.set_cookie(SESSION_COOKIE, sessionId,
                                    secure=https, httponly=True)
                response.headers[SESSION_HEADER] = sessionId
            elif SESSION_COOKIE in request.cookies:
                response.set_cookie(SESSION_COOKIE, None)
        finally:
            request.client.commit()
            del request.client

    @extract_params('username?', 'password?', 'otp?')
    def authenticate(self, request, username=None, password=None, otp=None):
        try:
            request.client.authenticate(username, password, otp)
            return json_response(True)
        except:
            return json_response(False, status=400)

    @extract_params('username?', 'password?', 'otp?')
    def login(self, request, username=None, password=None, otp=None):
        try:
            session = request.client.create_session(username, password, otp)
            request.session = session
            return json_response(True)
        except Exception as e:
            log.warn(e)
            if request.session:
                request.session.delete()
                request.session = None
            return json_error('Invalid credentials!')

    @require_session
    def logout(self, request):
        request.session.delete()
        request.session = None
        return json_response(True)

    @require_session
    def status(self, request):
        return json_response(request.session.data)

    @require_session
    @extract_params('oldpass', 'newpass', 'otp?')
    def change_password(self, request, oldpass, newpass, otp=None):
        user = request.session.user
        try:
            request.client.authenticate(user.name, oldpass, otp)
            user.set_password(newpass)
            return json_response(True)
        except:
            return json_error('Invalid credentials!')

    @require_session
    @extract_params('yubikey', 'password', 'otp?')
    def assign_yubikey(self, request, yubikey, password, otp=None):
        user = request.session.user
        try:
            request.client.authenticate(user.name, password, otp)
            prefix = yubikey[:-32]
            if not validate_otp(yubikey):
                return json_error('Invalid OTP for new YubiKey!')
            if not prefix in user.yubikeys:
                user.assign_yubikey(prefix)
            return json_response(True)
        except:
            return json_error('Invalid credentials!')

    @require_session
    @extract_params('otp')
    def generate_revocation(self, request, otp):
        user = request.session.user
        if not user.validate_otp(otp):
            return json_error('Invalid credentials!')
        code = request.client.generate_revocation(otp[:-32])
        return json_response(code)

    @extract_params('code')
    def revoke_yubikey(self, request, code):
        try:
            self.client.revoke(code)
            return json_response(True)
        except:
            return json_error('Invalid code!')


application = ClientAPI('/%s/client' % settings['rest_path'])

if __name__ == '__main__':
    httpd = make_server('localhost', 8080, application)
    httpd.serve_forever()



from passlib.context import CryptContext
from passlib.registry import register_crypt_handler_path

register_crypt_handler_path('yhsm_pbkdf2_sha1', 'yubiauth.yhsm')
register_crypt_handler_path('yhsm_pbkdf2_sha256', 'yubiauth.yhsm')
register_crypt_handler_path('yhsm_pbkdf2_sha512', 'yubiauth.yhsm')

from nose.plugins.attrib import attr

PASSWORDS = [
    'foobar',
    '',
    '1234567890',
    '!"#%&/()=?',
    chr(150) + chr(200) + chr(255)
]

context = None


@attr(hsm=True)
def setup():
    global context
    context = CryptContext(
        schemes=['yhsm_pbkdf2_sha1', 'yhsm_pbkdf2_sha256',
                 'yhsm_pbkdf2_sha512'],
        default='yhsm_pbkdf2_sha1',
        all__key_handle=1,
        all__rounds=10
    )


def _algorithm_test(scheme):
    for pwd in PASSWORDS:
        res = context.encrypt(pwd, scheme=scheme)
        assert context.identify(res) == scheme
        assert context.verify(pwd, res)
        assert res != context.encrypt(pwd, scheme=scheme)


@attr(hsm=True)
def test_yhsm_pbkdf2_sha1():
    _algorithm_test('yhsm_pbkdf2_sha1')


@attr(hsm=True)
def test_yhsm_pbkdf2_sha256():
    _algorithm_test('yhsm_pbkdf2_sha256')


@attr(hsm=True)
def test_yhsm_pbkdf2_sha512():
    _algorithm_test('yhsm_pbkdf2_sha512')



from nose import with_setup
from nose.tools import raises

from yubiauth import create_tables
from yubiauth.core import YubiAuth
from utils import setting

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

engine = create_engine('sqlite:///:memory:', echo=True)
create_tables(engine)
Session = sessionmaker(bind=engine)
auth = None


def setup():
    global auth
    auth = YubiAuth(Session())
    teardown()
    auth.create_user('user1', 'p4ssw0rd')
    auth.create_user('user2', 'foo')
    auth.commit()


def teardown():
    for user in auth.query_users():
        auth.get_user(user['id']).delete()
    auth.commit()


@with_setup(setup, teardown)
def test_create_users():
    user = auth.create_user('test_user', 'test_password')
    assert user.name == 'test_user'


@with_setup(setup, teardown)
@raises(Exception)
def test_create_existing_username():
    auth.create_user('user1', 'testing')


@with_setup(setup, teardown)
def test_validate_password():
    user = auth.get_user('user1')
    assert user.validate_password('p4ssw0rd')
    assert not user.validate_password('foo')
    assert not user.validate_password('bar')

    user2 = auth.get_user('user2')
    assert user2.validate_password('foo')
    assert not user2.validate_password('bar')


@with_setup(setup, teardown)
def test_empty_password():
    user = auth.get_user('user1')
    user.set_password(None)

    assert not user.validate_password(None)
    assert not user.validate_password('')

    with setting(allow_empty=True):
        assert user.validate_password(None)
        assert user.validate_password('')



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from wsgiref.simple_server import make_server
from webob import exc
import re

from yubiauth import settings
from yubiauth.core import YubiAuth
from yubiauth.util.rest import (
    REST_API, Route, no_content, json_response, json_error, extract_params)

ID_PATTERN = r'\d+'
USERNAME_PATTERN = r'(?=.*[a-zA-Z])[-_a-zA-Z0-9]{3,}'
PASSWORD_PATTERN = r'\S{3,}'
YUBIKEY_PATTERN = r'[cbdefghijklnrtuv]{0,64}'
ATTRIBUTE_KEY_PATTERN = r'[-_a-zA-Z0-9]+'

ID_RE = re.compile(r'^%s$' % ID_PATTERN)


class CoreAPI(REST_API):
    __user__ = r'^users/(%s|%s)' % (ID_PATTERN, USERNAME_PATTERN)
    __user_attribute__ = __user__ + r'/attributes/(%s)' % ATTRIBUTE_KEY_PATTERN
    __user_yubikey__ = __user__ + r'/yubikeys/(%s)' % YUBIKEY_PATTERN
    __yubikey__ = r'^yubikeys/(%s)' % YUBIKEY_PATTERN
    __yubikey_attribute__ = __yubikey__ + r'/attributes/(%s)' %\
        ATTRIBUTE_KEY_PATTERN
    __user_yubikey_attribute__ = __user_yubikey__ + r'/attributes/(%s)' %\
        ATTRIBUTE_KEY_PATTERN

    __routes__ = [
        Route(r'^user$', 'find_user'),
        Route(r'^users$', get='list_users', post='create_user'),
        Route(__user__ + r'$', get='show_user', delete='delete_user'),
        Route(__user__ + r'/reset$', post='reset_password'),
        Route(__user__ + r'/delete$', post='delete_user'),
        Route(__user__ + r'/validate$', 'validate'),
        Route(__user__ + r'/attributes$', get='list_user_attributes',
              post='set_user_attribute'),
        Route(__user_attribute__ + r'$', get='show_user_attribute',
              delete='unset_user_attribute'),
        Route(__user_attribute__ + r'/delete$', post='unset_user_attribute'),
        Route(__user__ + r'/yubikeys$', get='list_yubikeys',
              post='bind_yubikey'),

        Route(__user_yubikey__ + r'$', get='show_yubikey',
              delete='unbind_yubikey'),
        Route(__user_yubikey__ + r'/delete$', post='unbind_yubikey'),
        Route(__user_yubikey__ + r'/attributes$',
              get='list_yubikey_attributes', post='set_yubikey_attribute'),
        Route(__user_yubikey_attribute__ + r'$', get='show_yubikey_attribute',
              delete='unset_yubikey_attribute'),
        Route(__user_yubikey_attribute__ + r'/delete$',
              post='unset_yubikey_attribute'),

        Route(__yubikey__ + r'$', get='show_yubikey', delete='delete_yubikey'),
        Route(__yubikey__ + r'/delete$', post='delete_yubikey'),
        Route(__yubikey__ + r'/attributes$', get='list_yubikey_attributes',
              post='set_yubikey_attribute'),
        Route(__yubikey_attribute__ + r'$', get='show_yubikey_attribute',
              delete='unset_yubikey_attribute'),
        Route(__yubikey_attribute__ + r'/delete$',
              post='unset_yubikey_attribute')
    ]

    def _call_setup(self, request):
        request.auth = YubiAuth()

    def _call_teardown(self, request, response):
        request.auth.commit()
        del request.auth

    def _get_user(self, request, username_or_id):
        if ID_RE.match(username_or_id):
            username_or_id = int(username_or_id)

        try:
            return request.auth.get_user(username_or_id)
        except:
            raise exc.HTTPNotFound

    def _get_yubikey(self, request, *args):
        try:
            if len(args) == 1:
                return request.auth.get_yubikey(args[0])
            elif len(args) >= 2:
                user = self._get_user(request, args[0])
                return user.yubikeys[args[1]]
        except:
            pass

        raise exc.HTTPNotFound

    # Users

    def find_user(self, request):
        users = request.auth.query_users(**request.params)
        if len(users) == 1:
            user_id = users[0]['id']
            user = request.auth.get_user(user_id)
            response = json_response(user.data)
            response.headers.add('Link', '<%s>; rel="canonical"' %
                                 request.relative_url('users/%d' % user_id))
            return response

        raise exc.HTTPNotFound

    def list_users(self, request):
        return json_response(request.auth.query_users(**request.params))

    @extract_params('username', 'password')
    def create_user(self, request, username, password):
        try:
            user = request.auth.create_user(username, password)
            request.auth.commit()
            url = '%s/users/%d' % (self._base_path, user.id)
            return json_response({
                'id': user.id,
                'name': user.name
            }, location=url, status=201)
        except Exception, e:
            return json_error(e.message)

    def show_user(self, request, username_or_id):
        user = self._get_user(request, username_or_id)
        return json_response(user.data)

    @extract_params('password')
    def reset_password(self, request, username_or_id, password):
        user = self._get_user(request, username_or_id)
        user.set_password(password)

        return no_content()

    def delete_user(self, request, username_or_id):
        user = self._get_user(request, username_or_id)
        user.delete()

        return no_content()

    # Attributes

    def _list_attributes(self, owner):
        return json_response(owner.attributes.copy())

    def list_user_attributes(self, request, username_or_id):
        return self._list_attributes(self._get_user(request, username_or_id))

    def list_yubikey_attributes(self, request, *args):
        return self._list_attributes(self._get_yubikey(request, *args))

    def _show_attribute(self, owner, attribute_key):
        if attribute_key in owner.attributes:
            return json_response(owner.attributes[attribute_key])
        return json_response(None)

    def show_user_attribute(self, request, username_or_id, attribute_key):
        return self._show_attribute(self._get_user(request, username_or_id),
                                    attribute_key)

    def show_yubikey_attribute(self, request, *args):
        attribute_key = args[-1]
        return self._show_attribute(self._get_yubikey(request, *args[:-1]),
                                    attribute_key)

    @extract_params('key', 'value')
    def _set_attribute(self, request, owner, key, value):
        owner.attributes[key] = value

        return no_content()

    def set_user_attribute(self, request, username_or_id):
        return self._set_attribute(request, self._get_user(request,
                                                           username_or_id))

    def set_yubikey_attribute(self, request, *args):
        return self._set_attribute(request, self._get_yubikey(request, *args))

    def _unset_attribute(self, owner, attribute_key):
        if attribute_key in owner.attributes:
            del owner.attributes[attribute_key]

        return no_content()

    def unset_user_attribute(self, request, username_or_id, attribute_key):
        return self._unset_attribute(self._get_user(request, username_or_id),
                                     attribute_key)

    def unset_yubikey_attribute(self, request, *args):
        attribute_key = args[-1]
        return self._unset_attribute(self._get_yubikey(request, *args[:-1]),
                                     attribute_key)

    # YubiKeys

    def list_yubikeys(self, request, username_or_id):
        user = self._get_user(request, username_or_id)
        return json_response(user.yubikeys.keys())

    def show_yubikey(self, request, *args):
        yubikey = self._get_yubikey(request, *args)
        return json_response(yubikey.data)

    @extract_params('yubikey')
    def bind_yubikey(self, request, username_or_id, yubikey):
        user = self._get_user(request, username_or_id)
        user.assign_yubikey(yubikey)

        return no_content()

    def unbind_yubikey(self, request, username_or_id, prefix):
        user = self._get_user(request, username_or_id)
        del user.yubikeys[prefix]

        return no_content()

    def delete_yubikey(self, request, prefix):
        yubikey = self._get_yubikey(request, prefix)
        yubikey.delete()

        return no_content()

    # Validate

    def validate(self, request, username_or_id):
        user = self._get_user(request, username_or_id)

        if 'password' in request.params:
            password = request.params['password']
            valid_pass = user.validate_password(password)
        else:
            valid_pass = False

        if 'otp' in request.params:
            otp = request.params['otp']
            valid_otp = user.validate_otp(otp)
        else:
            valid_otp = False

        return json_response({
            'valid_password': valid_pass,
            'valid_otp': valid_otp
        })


application = CoreAPI('/%s/core' % settings['rest_path'])


if __name__ == '__main__':
    httpd = make_server('localhost', 8080, application)
    httpd.serve_forever()



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'yhsm_pbkdf2_sha1',
    'yhsm_pbkdf2_sha256',
    'yhsm_pbkdf2_sha512'
]

import os
from base64 import b64encode, b64decode
from passlib.utils import (to_hash_str, to_unicode)
from passlib.hash import pbkdf2_sha1, pbkdf2_sha256, pbkdf2_sha512

from pyhsm.base import YHSM
from pyhsm.util import key_handle_to_int

_UDOLLAR = u'$'
_UKH = u'kh='
_UDEFAULT_KH = u'1'

DEFAULT_DEVICE = 'daemon://localhost:5348'

hsm = YHSM(device=os.getenv('YHSM_DEVICE', DEFAULT_DEVICE))


def _yhsm__init__(base, self, key_handle=_UDEFAULT_KH, **kwds):
    super(base, self).__init__(**kwds)
    self.key_handle = key_handle


def _yhsmfrom_string(base, cls, hash):
    if not hash:
        raise ValueError('No hash specified')
    hash = to_unicode(hash, 'ascii', 'hash')

    if not hash.startswith(cls.ident):
        raise ValueError('invalid %s hash' % (cls.name))

    hash = hash[len(cls.ident):]

    if hash.startswith(_UKH):
        part, hash = hash.split(_UDOLLAR, 1)
        key_handle = part[len(_UKH):]
    else:
        key_handle = _UDEFAULT_KH

    inner_config, chk = hash.rsplit('$', 1)
    inner = base.from_string('%s%s' % (base.ident, inner_config))

    params = {}
    for kwd in inner.setting_kwds:
        try:
            params[kwd] = inner.__getattribute__(kwd)
        except AttributeError:
            pass
    params['key_handle'] = key_handle
    params['checksum'] = b64decode(chk.encode('ascii'))

    return cls(**params)


def _yhsm_hash_needs_update(base, cls, hash, **opts):
    key = 'deprecated_key_handles'
    if key in opts:
        for kh in opts[key].split(','):
            if 'kh=%s' % kh in hash:
                return True
    return False


def _yhsmto_string(base, self):
    hash = self.ident

    inner_str = super(base, self).to_string()
    inner_str = inner_str[len(self.ident):].rsplit('$', 1)[0]

    chk = b64encode(self.checksum).decode('ascii')

    hash += u'%s%s$%s$%s' % (
        _UKH,
        self.key_handle,
        inner_str,
        chk
    )

    return to_hash_str(hash)


def _yhsmcalc_checksum(base, self, secret):
    base_chk = super(base, self).calc_checksum(secret)
    result = hsm.hmac_sha1(key_handle_to_int(self.key_handle), base_chk)

    return result.result.hash_result


def _make_yhsm_handler(base, base_name):
    name = 'yhsm_%s' % (base_name)
    ident = u'$yhsm-%s$' % (base_name.replace('_', '-'))
    return type(name, (base,), dict(
        name=name,
        ident=ident,
        setting_kwds=('key_handle',) + base.setting_kwds,
        checksum_size=20,
        __init__=lambda *args, **kwargs: _yhsm__init__(base, *args, **kwargs),
        from_string=classmethod(lambda *args, **kwargs: _yhsmfrom_string(
            base, *args, **kwargs)),
        _hash_needs_update=classmethod(lambda *args, **kwargs:
                                       _yhsm_hash_needs_update(
                                           base, *args, **kwargs)),
        to_string=lambda *args, **kwargs: _yhsmto_string(
            base, *args, **kwargs),
        calc_checksum=lambda *args, **kwargs: _yhsmcalc_checksum(base, *args,
                                                                 **kwargs)
    ))


yhsm_pbkdf2_sha1 = _make_yhsm_handler(pbkdf2_sha1, 'pbkdf2_sha1')
yhsm_pbkdf2_sha256 = _make_yhsm_handler(pbkdf2_sha256, 'pbkdf2_sha256')
yhsm_pbkdf2_sha512 = _make_yhsm_handler(pbkdf2_sha512, 'pbkdf2_sha512')



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from yubiauth.util.controller import Controller
from yubiauth.util.model import Session
from yubiauth.core.model import User, YubiKey, AttributeAssociation

__all__ = [
    'YubiAuth'
]


class YubiAuth(Controller):
    """
    Main class for interacting with the YubiAuth backend.
    """
    def __init__(self, session=Session()):
        super(YubiAuth, self).__init__(session)

    def query_users(self, **kwargs):
        """
        Performs a query on all available users.

        Gets a list of all users matching the filter, represented as dicts
        containing id and name.

        Filtering is dony by supplying keyword arguments, where each key-value
        pair will match an Attribute for the user.

        A special keyword "yubikey" will create a filter on users assigned
        to the YubiKey with the prefix of the value given.

        Example:

        # Get users with the YubiKey ccccccccccce:
        query_users(yubikey='ccccccccccce')

        # Get users with the attribute 'area' equal to 'Stockholm'
        # AND the attribute 'admin' equal to 'True':
        query_users(area='Stockholm', admin='True')

        For more advanced querying, use the session attribute directly.

        @return: A list of users
        @rtype: list
        """
        query = self.session.query(User.id, User.name)

        if 'yubikey' in kwargs:
            query = query.filter(User.yubikeys.any(prefix=kwargs['yubikey']))
            del kwargs['yubikey']

        for key in kwargs:
            query = query.filter(
                User._attribute_association.has(
                    AttributeAssociation._attributes.any(
                        key=key,
                        value=kwargs[key]
                    )
                )
            )

        return [
            {'id': row[0], 'name': row[1]}
            for row in query.all()
        ]

    def query_yubikeys(self, **kwargs):
        """
        Performs a query on all available YubiKeys.

        Gets a list of all YubiKeys matching the filter.

        Filtering is dony by supplying keyword arguments, where each key-value
        pair will match an Attribute for the YubiKey.

        Example:
        # Get YubiKey with the attribute 'revoke' equal to 'foo':
        query_yubikeys(revoke='foo')

        @return: A list of YubiKeys
        @rtype: list
        """
        query = self.session.query(YubiKey)

        for key in kwargs:
            query = query.filter(
                YubiKey._attribute_association.has(
                    AttributeAssociation._attributes.any(
                        key=key,
                        value=kwargs[key]
                    )
                )
            )

        return query.all()

    def get_user(self, user_username_or_id):
        """
        Does a lookup for a user based on a username or ID.

        For practicality also checks if the argument is a C{User},
        in which case it is returned as-is.

        @param user_username_or_id: A username or user ID
        @type user_username_or_id: C{User} or string or int
        """
        if isinstance(user_username_or_id, User):
            return user_username_or_id
        else:
            query = self.session.query(User)
            try:
                if isinstance(user_username_or_id, int):
                    user = query.get(user_username_or_id)
                    if user:
                        return user
                else:
                    return query.filter(User.name == user_username_or_id).one()
            except:
                pass

        raise LookupError('User not found!')

    def get_yubikey(self, prefix):
        """
        Gets a YubiKey by its prefix.

        @param prefix: A YubiKey prefix
        @type prefix: string
        """
        return self.session.query(YubiKey).filter(
            YubiKey.prefix == prefix).one()

    def create_user(self, username, password):
        """
        Creates a new user.

        @param username: A unique username to give the new user.
        @type username: string

        @param password: The password to give the new user.
        @type password: string

        @return: The created user.
        @rtype: C{User}
        """
        try:
            self.get_user(username)
        except:
            user = User(username, password)
            self.session.add(user)
            return user
        raise ValueError('A user with that username already exists!')



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'engine',
    'Session',
    'Deletable'
]

from yubiauth.config import settings

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


engine = create_engine(settings['db'], echo=False)
Session = sessionmaker(bind=engine)


class Deletable(object):
    """
    Mixin for model objects which should have a delete method.
    """
    def delete(self):
        session = Session.object_session(self)
        if not session:
            pass
        elif self in session.new:
            session.expunge(self)
        else:
            session.delete(self)



from webtest import TestApp
from yubiauth.core.rest import application
from yubiauth.util.model import engine
from yubiauth import create_tables, YubiAuth


create_tables(engine)

app = TestApp(application)


# Users
def setup():
    auth = YubiAuth()
    for user in auth.query_users():
        auth.get_user(user['id']).delete()
    auth.commit()
    del auth


def test_create_user():
    resp = app.post(
        '/yubiauth/core/users',
        {'username': 'user1', 'password': 'foo'},
        status=201
    )
    user = app.get(resp.location).json
    assert user['name'] == 'user1'
    assert user['id'] == 1
    app.post('/yubiauth/core/users', {'username': 'user2', 'password': 'bar'},
             status=201)
    app.post('/yubiauth/core/users', {'username': 'user3', 'password': 'baz'},
             status=201)

    assert len(app.get('/yubiauth/core/users').json) == 3


def test_delete_user_by_id():
    resp = app.post(
        '/yubiauth/core/users',
        {'username': 'redshirt', 'password': 'llap'},
        status=201
    )
    id = app.get(resp.location).json['id']

    app.delete('/yubiauth/core/users/%d' % id)
    app.get('/yubiauth/core/users/%d' % id, status=404)


def test_delete_user_by_name():
    resp = app.post(
        '/yubiauth/core/users',
        {'username': 'redshirt', 'password': 'llap'},
        status=201
    )
    name = app.get(resp.location).json['name']

    app.delete('/yubiauth/core/users/%s' % name.encode('ascii'))
    app.get('/yubiauth/core/users/%s' % name, status=404)


def test_delete_user_by_post():
    resp = app.post(
        '/yubiauth/core/users',
        {'username': 'redshirt', 'password': 'llap'},
        status=201
    )
    id = app.get(resp.location).json['id']

    app.post('/yubiauth/core/users/%d/delete' % id)
    app.get('/yubiauth/core/users/%d' % id, status=404)


def test_get_user_by_id():
    user = app.get('/yubiauth/core/users/1', status=200).json
    assert user['id'] == 1
    assert user['name'] == 'user1'


def test_get_user_by_name():
    user = app.get('/yubiauth/core/users/user1').json
    assert user['id'] == 1
    assert user['name'] == 'user1'


def test_validate_password_get():
    status = app.get('/yubiauth/core/users/user1/validate?password=foo'
                     ).json
    assert status['valid_password']


def test_validate_password_post():
    status = app.post(
        '/yubiauth/core/users/user1/validate',
        {'password': 'foo'}
    ).json
    assert status['valid_password']


def test_reset_password():
    status = app.get('/yubiauth/core/users/user1/validate?password=foo').json
    assert status['valid_password']
    app.post('/yubiauth/core/users/1/reset', {'password': 'foobar'})

    status = app.get('/yubiauth/core/users/user1/validate?password=foo').json
    assert not status['valid_password']

    status = app.get('/yubiauth/core/users/user1/validate?password=foobar'
                     ).json
    assert status['valid_password']


def test_create_user_with_existing_username():
    app.post('/yubiauth/core/users', {
             'username': 'user1', 'password': 'bar'}, status=400)


def test_validate_with_invalid_username():
    app.post('/yubiauth/core/users/notauser/validate',
             {'password': 'foo'}, status=404)

    app.post('/yubiauth/core/users/notauser/validate', status=404)


def test_validate_with_invalid_password():
    status = app.post('/yubiauth/core/users/user1/validate',
                      {'password': 'wrongpassword'}).json
    assert not status['valid_password']

    status = app.post('/yubiauth/core/users/user1/validate').json
    assert not status['valid_password']


def test_get_user_by_invalid_username():
    assert app.get('/yubiauth/core/users/notauser', status=404)


# YubiKeys


PREFIX_1 = 'ccccccccccce'
PREFIX_2 = 'cccccccccccd'
PREFIX_3 = 'cccccccccccf'


def test_bind_yubikeys():
    yubikeys = app.get('/yubiauth/core/users/1/yubikeys').json
    assert len(yubikeys) == 0

    app.post('/yubiauth/core/users/1/yubikeys', {'yubikey': PREFIX_1})
    yubikeys = app.get('/yubiauth/core/users/1/yubikeys').json
    assert yubikeys == [PREFIX_1]

    app.post('/yubiauth/core/users/1/yubikeys', {'yubikey': PREFIX_2})
    app.post('/yubiauth/core/users/2/yubikeys', {'yubikey': PREFIX_2})

    yubikeys = app.get('/yubiauth/core/users/1/yubikeys').json
    assert sorted(yubikeys) == sorted([PREFIX_1, PREFIX_2])

    user = app.get('/yubiauth/core/users/1').json
    assert sorted(user['yubikeys']) == sorted([PREFIX_1, PREFIX_2])

    user = app.get('/yubiauth/core/users/2').json
    assert user['yubikeys'] == [PREFIX_2]


def test_show_yubikey():
    yubikey = app.get('/yubiauth/core/users/1/yubikeys/%s' % PREFIX_1).json
    assert yubikey['enabled']
    assert yubikey['prefix'] == PREFIX_1


def test_show_yubikey_for_wrong_user():
    app.get('/yubiauth/core/users/2/yubikeys/%s' % PREFIX_1, status=404)


def test_unbind_yubikeys():
    app.delete('/yubiauth/core/users/1/yubikeys/%s' % PREFIX_1)
    yubikeys = app.get('/yubiauth/core/users/1/yubikeys').json
    assert yubikeys == [PREFIX_2]


def test_find_user_by_yubikey():
    app.post('/yubiauth/core/users/2/yubikeys', {'yubikey': PREFIX_3})

    user = app.get('/yubiauth/core/user?yubikey=%s' % PREFIX_3).json
    assert user['id'] == 2
    assert PREFIX_3 in user['yubikeys']


# Attributes


def test_assign_attributes():
    app.post('/yubiauth/core/users/1/attributes', {
             'key': 'attr1', 'value': 'val1'})
    app.post('/yubiauth/core/users/1/attributes', {
             'key': 'attr2', 'value': 'val2'})

    attributes = app.get('/yubiauth/core/users/1/attributes').json
    assert attributes['attr1'] == 'val1'
    assert attributes['attr2'] == 'val2'
    assert len(attributes) == 2

    user = app.get('/yubiauth/core/users/1').json
    assert user['attributes'] == attributes


def test_find_user_by_attribute():
    user1 = app.get('/yubiauth/core/user?attr1=val1&attr2=val2').json
    assert user1['id'] == 1

    user2 = app.post('/yubiauth/core/user', {
                     'attr1': 'val1', 'attr2': 'val2'}).json
    assert user1 == user2


def test_read_attribute():
    value = app.get('/yubiauth/core/users/1/attributes/attr1').json
    assert value == 'val1'


def test_read_missing_attribute():
    value = app.get('/yubiauth/core/users/1/attributes/foo').json
    assert not value


def test_overwrite_attributes():
    app.post('/yubiauth/core/users/1/attributes', {
             'key': 'attr1', 'value': 'newval'})
    attributes = app.get('/yubiauth/core/users/1/attributes').json

    assert attributes['attr1'] == 'newval'
    assert attributes['attr2'] == 'val2'
    assert len(attributes) == 2


def test_unset_attributes():
    app.delete('/yubiauth/core/users/1/attributes/attr1')
    attributes = app.get('/yubiauth/core/users/1/attributes').json

    assert attributes['attr2'] == 'val2'
    assert len(attributes) == 1

    value = app.get('/yubiauth/core/users/1/attributes/attr1').json
    assert not value


def _basic_attribute_test(base_url):
    values = {
        'attr1': 'foo',
        'attr2': 'bar',
        'attr3': 'baz',
    }

    for key, value in values.items():
        app.post(base_url, {'key': key, 'value': value})
        assert app.get('%s/%s' % (base_url, key)).json == value

    app.delete('%s/attr2' % (base_url))
    assert not app.get('%s/attr2' % base_url).json

    app.post(base_url, {'key': 'attr3', 'value': 'newval'})
    assert app.get('%s/attr3' % base_url).json == 'newval'

    del values['attr2']
    values['attr3'] = 'newval'
    assert app.get(base_url).json == values


def test_user_attributes():
    _basic_attribute_test('/yubiauth/core/users/1/attributes')
    _basic_attribute_test('/yubiauth/core/users/2/attributes')


def test_yubikey_attributes():
    _basic_attribute_test('/yubiauth/core/yubikeys/%s/attributes' % PREFIX_1)
    _basic_attribute_test(
        '/yubiauth/core/users/2/yubikeys/%s/attributes' % PREFIX_2)
